/*
 * UNPUBLISHED PROPRIETARY INFORMATION
 * Copyright (c) 2007-2008 Sentilla Corporation
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Sentilla Corporation ("Confidential Information"). Disclosure of this
 * Confidential Information is prohibited, and its use shall be in
 * compliance with terms of the license agreement with Sentilla Corporation.
 */


import javax.measure.quantity.ElectricPotential;

import com.sentilla.platform.tmote.AccelerationSensor;
import com.sentilla.platform.tmote.VoltageAdc;
import com.sentilla.system.ActionException;
import com.sentilla.system.DigitalPinDriver;
import com.sentilla.system.Sensor;

/**
 * Sample Accelerometer driver: takes care of initializing, calibration, and reading.
 * It uses the multiple channels reading mode of the ADC. It reads the 
 * Vref every 10 minutes.
 * 
 * @author Sentilla Corporation
 *
 */
public class Accelerometer
{
	public boolean isAsleep;
	public boolean XOn;
	public boolean YOn;
	public boolean ZOn;
	public double sensitivity; 
	private Sensor<ElectricPotential> AccSensor;
	private byte[] portsToRead;
	private int numPorts;
	private float Vref;
	private final static int OBSOLETE_VREF = 600000;
	private long lastTimeVrefWasChecked;
	private float[] calibration;
    private int[] raw;
	
    /**
     * Default constructor uses all three axes.
     *
     */
	public Accelerometer()
	{
        this(true, true, true);
	}
	
    /**
     * Construct the accelerometer object to query only the desired
     * axes.
     * 
     * @param x true if polling xaxis
     * @param y true if polling yaxis
     * @param z true if polling yaxis
     */
	public Accelerometer(boolean x, boolean y, boolean z) {
        // Set the last time Vref checked to the max (10 minutes)
		lastTimeVrefWasChecked = -OBSOLETE_VREF - 1;

        // Instantiate the voltageAdc object for reading the accelerometer pins
		AccSensor = new VoltageAdc(VoltageAdc.PORT_A4, VoltageAdc.REFVOLT_2_5);
	
        // Initially, the pins are set as output.  Reading into them will set them
        // as inputs
        DigitalPinDriver.create(0x64).read();
        DigitalPinDriver.create(0x65).read();
        DigitalPinDriver.create(0x66).read();
        
        // Set the axis that will be polled by the Accelerometer class
		this.setAxis(x, y, z);
		
        // Turn on the accelerometer
		this.setSleep(false);
        
        // Set the sensitivity level to the lowest setting
        this.setGRange(AccelerationSensor.G_RANGE_1_5);
        
        // Calibrate to the initial mote orientation
		calibrate();
	}
	
	/**
     * Calibrate the accelerometer values to 0 based on the initial
     * startup orientation of the mote. 
     *
	 */
	public void calibrate()
	{
		short index = 0;
		calibration = new float[numPorts];
        
        // Get the reference voltage
		getVref();
        
		// Read the initial values from the accelerometer
        ((com.sentilla.platform.tmote.Adc<ElectricPotential>)AccSensor).readMultipleRaw(raw, portsToRead);
        
        // Convert the fixed point decimal values to floats
        for (int i = 0; i<numPorts; i++)
            calibration[i] = (float) (raw[i]/(1024.0*1024.0));
        
        // Set the calibration array to the initial values
		if (XOn)
			calibration[index] = - (float)((calibration[index++] - Vref) / sensitivity);
		if (YOn)
			calibration[index] = - (float)((calibration[index++] - Vref) / sensitivity);
		if (ZOn)
			calibration[index] = - ((float)((calibration[index++] - Vref) / sensitivity) - 1);
	}

	/**
     * Sets the sensitivity level of the accelerometer.  Refer to the data sheet
     * for more information about sensitivity.
     *  
     * @param range sensitivity in G-Force.
	 */
	public void setGRange(byte range)
	{
        // G force sensitivity is set with these two pins
        DigitalPinDriver.create(0x20).write( (range&1) != 0 );
        DigitalPinDriver.create(0x21).write( (range&2) != 0 );

        // Need the sensitivity level in later calculations
        switch (range) {
        case AccelerationSensor.G_RANGE_1_5:
            sensitivity = 0.8;
            break;
        case AccelerationSensor.G_RANGE_2:
            sensitivity = 0.6;
            break;
        case AccelerationSensor.G_RANGE_4:
            sensitivity = 0.3;
            break;
        case AccelerationSensor.G_RANGE_6:
            sensitivity = 0.2;
            break;
        default:
            throw new ActionException(); 
        }
	}
	
	
	/**
	 * Sets the accelerometer to sleep if input is true, wakes up if input
     * is false.
     * 
	 * @param sleep puts accelerometer to sleep if true, otherwise wakes up.
	 */
	public void setSleep(boolean sleep)
	{
        DigitalPinDriver.create(0x23).write(!sleep);
    	isAsleep = sleep;
	}

	
	/**
	 * Set the axes to read from the accelerometer.
	 * 
	 * @param x set x axis read.
	 * @param y set y axis read.
	 * @param z set z axis read.
	 */
	public void setAxis(boolean x, boolean y, boolean z)
	{
        // Throw an exception if trying to configure to not use any
        // axes
		if ((!x) && (!y) && (!z))
			throw new ActionException();
		
		if (AccSensor == null)	// Should never happen
			AccSensor = new VoltageAdc(VoltageAdc.PORT_A4, VoltageAdc.REFVOLT_2_5);

        // Set the axes that will be read by setting the correct ports
		portsToRead = new byte[((x ? 1 : 0) + (y ? 1 : 0) + (z ? 1 : 0))];
		numPorts = 0;
		XOn = x;
		YOn = y;
		ZOn = z;
		if (x)
			portsToRead[numPorts++] = VoltageAdc.PORT_A4;
		if (y)
			portsToRead[numPorts++] = VoltageAdc.PORT_A5;
		if (z)
			portsToRead[numPorts++] = VoltageAdc.PORT_A6;
        
        // Create the array for raw values from the accelerometer before converting
        // them to floats
        raw = new int[numPorts];
		return;
	}

	/**
     * Read the accelerometer values and place into the input array.  This
     * function will query from 1 to 3 axes and return the current acceleration
     * reported by the mote.  
     *  
     * @param readings
	 */
    public void getMultipleReadings( float[] readings ) {
        // Throw an exception is the accelerometer is asleep or not set
        // to receive on any axes
		if ((isAsleep) || ((!XOn) && (!YOn) && (!ZOn)))
            throw new ActionException();

        // Get the reference voltage.  This only actually submits to the kernel
        // once every ten minutes
		getVref();
        
        // Poll the accelerometer for the current values
		((com.sentilla.platform.tmote.Adc<ElectricPotential>)AccSensor).readMultipleRaw(raw, portsToRead);

        // Read the values and intepret them into g forces
        short index = 0;
		if (XOn)
            readings[index] = (float)(( (raw[index]/(1024.0 * 1024.0)) - Vref) / sensitivity) + calibration[index++];
		if (YOn)
            readings[index] = (float)(((raw[index]/(1024.0 * 1024.0)) - Vref) / sensitivity) + calibration[index++];
		if (ZOn)
            readings[index] = (float)(((raw[index]/(1024.0 * 1024.0)) - Vref) / sensitivity) + calibration[index++];
	}
	
    /**
     * Get the reference voltage of the accelerometer.  Use this for scaling the 
     * returned acceleration measurements.  This is set to only query once every
     * 10 mins to increase speed.
     * 
     * @return the reference voltage of the accelerometer.
     */
	public float getVref() {
		long currentTime = java.lang.System.currentTimeMillis();
        
        // only check the ref voltage if 10 mins have passed
		if ((currentTime - lastTimeVrefWasChecked) > OBSOLETE_VREF) {
            // Check the power port for the accelerometer
			byte[] port = new byte[] {VoltageAdc.PORT_AVCC_DIV_2};
			int[] result = new int[1];
			((com.sentilla.platform.tmote.Adc<ElectricPotential>)AccSensor).readMultipleRaw(result, port);
			Vref =  (float) (result[0]/(1024.0 * 1024.0));
			lastTimeVrefWasChecked = currentTime;
		}
		return(Vref);
	}
}

