/*
 * UNPUBLISHED PROPRIETARY INFORMATION
 * Copyright (c) 2007-2008 Sentilla Corporation
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Sentilla Corporation ("Confidential Information"). Disclosure of this
 * Confidential Information is prohibited, and its use shall be in
 * compliance with terms of the license agreement with Sentilla Corporation.
 */
import java.io.Serializable;

import com.sentilla.io.ByteBuffer;
import com.sentilla.net.Addressable;
import com.sentilla.net.BroadcastAddress;
import com.sentilla.net.ByteSenderDriver;
import com.sentilla.net.Mac64Address;
import com.sentilla.net.ReceiverDriver;
import com.sentilla.net.SenderDriver;
import com.sentilla.net.UnsupportedAddressException;
import com.sentilla.platform.tmote.Dispatcher;
import com.sentilla.platform.tmote.Protocol;
import com.sentilla.platform.tmote.Receiver;
import com.sentilla.platform.tmote.Sender;
import com.sentilla.system.Action;
import com.sentilla.system.Drivers;
import com.sentilla.system.Leds;
import com.sentilla.system.LedsDriver;

/**
 * Simple reliable transport protocol.
 * 
 * This application should only be run with two motes and a gateway.
 *
 * Sends a message and waits for an ack. If the ack is not received, it waits
 * 3 seconds (default) and sends the message again. It retries sending the message 4
 * times before failing. The receiver dispatches received messages to the user
 * application and sends an ack back to the sender.
 * 
 * <p>
 * During send and receive, one device acts as sender while the other receives.
 * Every ten seconds It will send a message containing a sequence number. The 
 * receiving device will activate LEDs to show the sequence number.
 * </p>
 * 
 * 
 * @author <a href="mailto:info@sentilla.com">Sentilla Corporation</a>
 */
public class MoteReliableProtocol extends Protocol {

    // Set the maximum size of a radio message for this protocol
    public final static int MTU = 64;

    public static final int SEND_TIMEOUT = 3000;

    public static final int SEND_RETRIES = 4;

    public static final int SEND_INTERVAL = 10000;

    private final static short LED_RECEIVE_READY = 1;

    private final static short LED_SEND_READY = 2;

    // ByteSender used to send ByteBuffers.  The ByteBuffer will
    // contain the serialized message object.
    private com.sentilla.net.ByteSender radioSender;

    // Keep a reference to the previous sender in case an ack is
    // received.
    private RTSender senderPtr;

    // Use a ByteBuffer to store serialized message objects.
    protected ByteBuffer msgBuf;

    // Use a ByteBuffer to store the serialized ack messages.
    protected ByteBuffer ackBuf;

    // A dispatcher reference used for dispatching messages to
    // applications.
    public Dispatcher ds;

    // The mote ID.
    private static long id;

    /**
     * Constructor will initialize fields.
     */
    public MoteReliableProtocol() {

        // Set the mote ID.
        MoteReliableProtocol.id = Mac64Address.getLocalAddress().longValue(); 

        // Create a ByteSender.  Use port 249 as the default for this protocol.
        // Send messages to the broadcast address.
        radioSender = ByteSenderDriver.create(249);

        // Initialize the fields.
        this.senderPtr = null;
        this.msgBuf = new ByteBuffer(MTU);
        this.ackBuf = new  ByteBuffer(MTU);

        // Use the Drivers method to get a reference to the Dispatcher.
        // This will be used to dispatch messages to user applications.
        ds = Drivers.getDriver("disp", Dispatcher.class);
    }

    /**
     * Method called when the Dispatcher receives a message for the MoteReliableProtocol.
     */
    public void stackReceive( Receiver rcvr ) {

        // Get the data object that was received by the communication interface.
        Object d = rcvr.getData();

        // Check if the message is of type RTMessage.  If so, then send an ack to
        // sending mote and dispatch the received message to user applications.
        if (d instanceof RTMessage) {
            // Cast the message correctly.
            RTMessage msg = (RTMessage)d;

            // Acknowledge reception of the message by sending an ACK.
            sendAck( msg.senderAddress );

            // Create a new RTReceiver object.
            RTReceiver r = new RTReceiver();

            // Add the previous receiver to the list of metadata in the
            // RTReceiver.
            r.addMetadata(rcvr);

            // Set the data field of the RTReceiver to the encapsulated data.
            r.data = msg.data;

            // Dispatch the newly created RTReceiver to the user applications.  This
            // will copy the received data to any user applications that are waiting
            // for that data.
            ds.dispatch(r);
        }
        // If this is an AckMessage, then signal the previous receiver
        // that the message has been acknowledged.
        else if (d instanceof AckMessage) {

            // Cast the message as an AckMessage.
            AckMessage msg = (AckMessage)d;

            // Check that the previous sender is non-null and that the AckMessage
            // address matches this motes address.
            if( (senderPtr != null)  && (msg.address == MoteReliableProtocol.id ) ) {

                // Set the sender success field to true.
                senderPtr.success = true;

                // Signal the sender that it is finished sending.
                senderPtr.setDone(Action.DONE);
            }
        }
    }

    /**
     * Called when sending a message with the MoteReliableProtocol.
     */
    public void stackSend( Sender ds ) {

        // Create a new RTMessage to be sent.  This message will encapsulate
        // the application message.
        RTMessage msg = new RTMessage();

        // Set the data field of the RTMessage to the application message.
        msg.data = ds.data;

        // Set the senderAddress field to this motes ID.
        msg.senderAddress = MoteReliableProtocol.id;

        // Cast the Sender object as an RTSender.
        RTSender s = (RTSender)ds;

        // The success field will indicate if the message has been successfully
        // transmitted.  It will be set true if an ACK is received.
        s.success = false;

        // Track what sender is currently waiting to receive an ACK.
        senderPtr = s;

        // Serialize the RTMessage for transmission.
        msgBuf.writeObject(msg);

        // Retry sending the message 4 times.
        for (int n=0; n<s.numRetries; n++) {

            // Send the serialized message over the radio.
            radioSender.send(msgBuf);

            // Sleep for awhile before resending the message.
            // This should give enough time for an ACK to be received.
            try {
                Thread.sleep(s.timeout);
            } catch( java.lang.InterruptedException e) {}

            // If an ACK has been received, stop transmitting the message.
            if (s.success == true)
                break;
        }

        // Clear the ByteBuffer for the next transmission.
        msgBuf.clear();

        // Signal the Sender object that it is done sending.
        s.setDone(Action.DONE);

        // Set the previous sender pointer to null.
        senderPtr = null;
    }

    /**
     * Send an ack message.
     */
    private void sendAck( long addr ) {

        // Create a new AckMessage.
        AckMessage msg = new AckMessage();

        // Set the address in the AckMessage to the sender of the
        // RTMessage.
        msg.address = addr;

        // Serialize the AckMessage into a ByteBuffer for transmission.
        ackBuf.writeObject(msg);

        // Send the AckMessage over the radio.
        radioSender.send(ackBuf);

        // Clear the AckMessage ByteBuffer of all data.  It is now ready
        // for the next transmission.
        ackBuf.clear();
    }

    /**
     * Create a new sender for the MoteReliableProtocol.
     */
    public Sender newSender() {
        return new RTSender();
    }

    /**
     * Create a new receiver for the MoteReliableProtocol.
     */
    public Receiver newReceiver() {
        return new Receiver();
    }

    /**
     * Returns true for all MoteReliableProtocol related messages.
     */
    public boolean doesInstanceMatch( Object obj ) {
        if( obj instanceof RTMessage) return true;
        if( obj instanceof AckMessage) return true;
        return false;
    }

    /**
     * Returns the string name associated with MoteReliabeProtocol.
     */
    public String getDriverName( int n ) {
        if (n==0) return "reliableprotocol";
        return null;
    }

    /**
     * Reliable protocol message class.
     */
    public static class RTMessage implements Serializable {

        // Should set the sending motes address in the message.
        public long senderAddress;

        // The encapsulated data that is being transmitted.
        public Object data;
    }

    /**
     * ACK message that will be sent by the MoteReliableProtocol.
     */
    public static class AckMessage implements Serializable {

        // Should contain the address of the mote that this ACK
        // is meant for.
        public long address;
    }

    /**
     * Sender class that contains sender metadata.
     */
    public static class RTSender extends Sender {

        // Metadata associated with sending a message.
        public int timeout;
        public int numRetries;
        public boolean success;

        /**
         * Set the default values for metadata.
         */
        public RTSender() {
            this.timeout = SEND_TIMEOUT;
            this.numRetries = SEND_RETRIES;
            this.success = false;
        }

        public void setAddress( Addressable addr ) {
            if (addr instanceof Mac64Address | addr instanceof BroadcastAddress)
                this.address = addr;
            else
                throw new UnsupportedAddressException();
        }
    }

    /**
     * Receiver class required but not needed by this protocol.
     */
    public static class RTReceiver extends Receiver {
    }


    /**
     * Test message to be used in testing the protocol.
     */
    public static class TestMessage implements Serializable {
        public int index = 0;
    }

    /**
     * One mote will send messages to the other motes on this thread.
     */
    public static class SendThread extends Thread {
        public boolean rcvdMessage = false;

        public void run() {
            Leds leds = LedsDriver.create();

            // Signal ready to send.
            leds.on(LED_SEND_READY);

            // Create a sample button using the accelerometer
            AcceleratorButton button = new AcceleratorButton();

            // Block until the mote shakes
            button.shake();

            // If already receiving message, then don't send any
            if (rcvdMessage)
                this.stop();

            // Create a Sender object using the MoteReliableProtocol.
            com.sentilla.net.Sender sender = SenderDriver.create("reliableprotocol");

            // Create a new message to send.
            TestMessage msg = new TestMessage();

            // Main sender execution loop.
            while(true) {
                try {
                    // Send the message.
                    sender.send(msg);

                    // If the transmission succeded, set the leds and increment
                    // the index.
                    if (((RTSender) sender).success == true) {
                        leds.set(msg.index);
                        msg.index++;
                    } else { // flash the LED for an error
                        for (short i = 0; i < 8; i++) {
                            leds.set(255);
                            Thread.sleep(100);
                        }
                    }

                    // Send one message every few seconds.
                    Thread.sleep(SEND_INTERVAL);

                } catch (InterruptedException e) { // ignore
                }
            }
        }

    }    

    /**
     * Test application for testing the functionality of the MoteReliableProtocol.
     */
    public static void motemain() throws InterruptedException {

        // Add the MoteReliableProtocol to the list of Drivers available
        // to the platform.
        Drivers.addDriver(new MoteReliableProtocol());

        // Create an LED object for setting the LEDs.
        Leds leds = LedsDriver.create();

        // Show a startup signal to indicate the app is alive.
        leds.set(255);
        Thread.sleep(2000);

        // Create a reference to a TestMessage and create a Receiver
        // object to receive a TestMessage.  The receiver will return
        // when a TestMessage is received.
        TestMessage msg;
        com.sentilla.net.Receiver receiver = ReceiverDriver.create(TestMessage.class);

        // Signal ready to receive.
        leds.set(LED_RECEIVE_READY);

        // Create and start send thread.
        // Start sending if the mote shakes.
        SendThread st = new SendThread();
        st.start();

        // Main receiver execution loop.
        while(true) {

            // Block until a message is received.
            msg = (TestMessage)receiver.receive();

            // Set the received flag to true so this mote doesn't start transmitting
            st.rcvdMessage = true; 

            leds.set(msg.index); // Set the LEDs to the msg index.
        }
    }
}
