/*
 * UNPUBLISHED PROPRIETARY INFORMATION
 * Copyright (c) 2007 - 2008 Sentilla Corporation
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Sentilla Corporation ("Confidential Information"). Disclosure of this
 * Confidential Information is prohibited, and its use shall be in
 * compliance with terms of the license agreement with Sentilla Corporation.
 */
import java.io.Serializable;

import com.sentilla.io.ByteBuffer;
import com.sentilla.net.Addressable;
import com.sentilla.net.BroadcastAddress;
import com.sentilla.net.ByteSenderDriver;
import com.sentilla.net.Mac64Address;
import com.sentilla.net.ReceiverDriver;
import com.sentilla.net.SenderDriver;
import com.sentilla.net.UnsupportedAddressException;
import com.sentilla.platform.tmote.Dispatcher;
import com.sentilla.platform.tmote.Protocol;
import com.sentilla.platform.tmote.Receiver;
import com.sentilla.platform.tmote.Sender;
import com.sentilla.system.Action;
import com.sentilla.system.Drivers;
import com.sentilla.system.Leds;
import com.sentilla.system.LedsDriver;

/**
 * Demonstrates a sample protocol that sends five copies of every message.
 * @author  <a href="mailto:info@sentilla.com">Sentilla Corporation</a>
 * <p>
 * Sends five copies of every message with a 1s delay between each copy. When
 * receiving, dispatches the first received copy of the message to user
 * applications and filters out all other received copies.
 * </p>
 * 
 * <p>
 * During Send and Receive, one device acts as Sender while the others receive.
 * Shaking a mote will designate it as the sending device. Every 10 seconds it 
 * will send a message containing a sequence number. The receiving devices will 
 * activate LEDs to show the sequence number.
 * </p>
 * 
 * @author Sentilla Corporation
 */
public class MoteTestProtocol extends Protocol {

    // Maximum size of transmitted data packets for the Sentilla point platform.
    private final static int MTU = 64;
    private final static short LED_RECEIVE_READY = 1;
    private final static short LED_SEND_READY = 2;

    // Static sequence number for tracking packets.
    private static int seqNum;

    // tracks whether this mote is receiving messages instead of sending them.
    private static boolean isReceiver = true;

    //Reference to the Dispatcher for dispatching messages to applicatons.
    private Dispatcher ds;

    // Reference to previous packet for filtering previously received messages.
    private TestMessage prevMessage;

    // ByteSender for sending ByteBuffers with the radio.
    private com.sentilla.net.ByteSender radioSender;

    // ByteBuffer used for sending messages.
    private com.sentilla.io.ByteBuffer buffer;

    // Mote ID.
    public long id;

    /**
     * Initialize the protocol members.
     */
    public MoteTestProtocol() {
        // Get the mode ID.
        id = Mac64Address.getLocalAddress().longValue();

        // Initialize the sequence number and prevMessage reference.
        seqNum = 0;
        prevMessage = null;

        // Get a ByteSender object for sending serialized messages.
        radioSender = ByteSenderDriver.create(14);

        // Initialize the ByteBuffer to the MTU size.
        buffer = new ByteBuffer(MTU);

        // Query the Drivers class for a reference to the Dispatcher.
        ds = (Dispatcher)Drivers.getDriver("disp", Dispatcher.class);
    }

    /**
     * Return a new sender for MoteTestProtocol.
     */
    public Sender newSender() {
        return new TestSender();
    }

    /**
     * Return a new receiver for MoteTestProtocol.
     */
    public Receiver newReceiver() {
        return new TestReceiver();
    }

    /**
     * Return true if the class of the input obj is TestMessage.
     */
    public boolean doesInstanceMatch( Object obj ) {
        return obj instanceof TestMessage;
    }

    /**
     * Called when the Dispatcher receives a message for this protocol.
     */
    public void stackReceive( Receiver rcvr ) {

        // Get the TestMessage contained in the Receiver object.
        TestMessage msg = (TestMessage) rcvr.getData();

        // Check hasn't been received already.
        if ( (prevMessage == null) || !msg.equals(prevMessage) ) {

            // Create a new TestReceiver to contain the received data
            // and the protocol metadata.
            TestReceiver meta = new TestReceiver();

            // Add the previous Receiver as metadata to the newly created
            // TestReceiver object.
            meta.addMetadata(rcvr);
            meta.data = msg.data;

            // Dispatch the new TestReceiver object.  This will send the
            // received data up to the user applications.
            ds.dispatch(meta);

            // Set the prevMessage reference to filter any other versions
            // of this message.
            prevMessage = msg;
        }
    }

    /**
     * Called when a message is sent with this protocol.
     */
    public void stackSend( Sender ds ) {
        // Cast the Sender as a TestSender.
        TestSender s = (TestSender)ds;

        // Create a new TestMessage to encapsulate the data to send.
        TestMessage msg = new TestMessage();

        // Set the encapsulated data to the Sender data (user data).
        msg.data = s.data;

        // Set the mote ID in the message.
        msg.address = id;

        // Set the sequence number for this message.
        msg.seqNum = seqNum++;

        // Set the port to send the messages to.
        radioSender.setPort(s.port);

        // Send the message 4 times.
        for (int i=4; i>=0; i--) {
            // Clear the ByteBuffer.
            buffer.clear();

            // Serialize the message into the ByteBuffer.
            buffer.writeObject(msg);

            // Send the ByteBuffer with the radio.
            radioSender.send(buffer);

            // Sleep for 1 seconds between each transmission.
            try {
                Thread.sleep(1000);
            } catch (java.lang.InterruptedException e) {}
        }

        // Signal the Sender object that the protocol is
        // finished sending the message.
        s.setDone(Action.DONE);
    }

    /**
     * Return the name of this protocol.
     */
    public String getDriverName( int n ) {
        if (n == 0) return "test";
        return null;
    }

    /**
     * Sender for MoteTestProtocol.  Contains metadata on the destination
     * address and port.
     */
    public static class TestSender extends Sender {
        public int port;

        /**
         * Initializes the TestSender object.
         */
        public TestSender() {
            // Note: all protocols must use port 249, or a application must be
            // created to receive the messages on a different port
            this.port = 249;
        }

        public void setAddress(Addressable addr) {
            if (addr instanceof BroadcastAddress)
                this.address = addr;
            else
                throw new UnsupportedAddressException();
        }
    }

    /**
     * Receiver for MoteTestProtocol.  This is required but not used.
     */
    public class TestReceiver extends Receiver {
    }

    /**
     * Message format for MoteTestProtocol.  Contains the mode ID,
     * a sequence number, and the data being sent.  Also includes
     * an equals method to make comparisons easier.
     */
    public static class TestMessage implements Serializable {

        public long address = -1;
        public int seqNum = 0;
        public Object data = null;

        /**
         * Return true if the input TestMessage is equal to this.
         */
        public boolean equals( TestMessage t ) {
            return ( (this.address == t.address) && (this.seqNum == t.seqNum) );
        }
    }


    /**
     * Driver application message.  This is the application message
     * that will be sent.
     */
    public static class MyMessage implements Serializable {
        public int index = 0;
    }

    /**
     * One mote will send messages to the other motes on this thread.
     */
    public static class SendThread extends Thread {
        public boolean rcvdMessage = false;

        public void run() {
            Leds leds = LedsDriver.create();
            
            // Signal ready to send.
            leds.on(LED_SEND_READY);
    
            // Create a sample button using the accelerometer
            AcceleratorButton button = new AcceleratorButton();

            // Block until the mote shakes
            button.shake();
            
            // If already receiving message, then don't send any
            if (rcvdMessage)
                this.stop();
            
            // Create a Sender object that uses the MoteTestProtocol.
            // Use the name of the protocol, "test", as the input
            // string to the protocol factory.
            com.sentilla.net.Sender sender = SenderDriver.create("test");

            // Create a new message to send.
            MyMessage msg = new MyMessage();

            // Main sender execution.
            while(true) {

                // Update a counter in the message.
                msg.index++;

                // Send the message.
                sender.send(msg);

                // Set the LEDs to the message count.
                leds.set(msg.index);

                // Sleep for 10 second between each transmission.
                try {
                    Thread.sleep(10000);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    /**
     * Application used to test the functionality of MoteTestProtocol.
     */
    public static void motemain() throws InterruptedException {

        // Add MoteTestProtocol as a new Driver available to the system.
        // This is required for all protocols.
        Drivers.addDriver(new MoteTestProtocol());

        // Create a new Leds object.
        Leds leds = LedsDriver.create();
        
        // Show a startup signal to indicate the app is alive.
        leds.set(255);
        Thread.sleep(2000);

        // Create a MyMessage variable.
        MyMessage msg = new MyMessage();

        // Create a Receiver object for the MyMessage class.  This
        // will return any messages received by the mote that match
        // the class MyMessage, regardless of how the message was
        // received.
        com.sentilla.net.Receiver receiver = ReceiverDriver.create(MyMessage.class);
        
        // Signal ready to receive.
        leds.set(LED_RECEIVE_READY);

        // Create and start send thread.
        // Start sending if the mote shakes.
        SendThread st = new SendThread();
        st.start();
        
        // Main receiver execution.
        while(isReceiver) {
            // Block until a message is received.
            msg = (MyMessage)receiver.receive();
            
            // Set the received flag to true so this mote doesn't start transmitting
            st.rcvdMessage = true;

            // Set the LEDs to the received message count.
            if (isReceiver)
                leds.set(msg.index);
        }
    }
}
