/*
 * UNPUBLISHED PROPRIETARY INFORMATION
 * Copyright (c) 2007-2008 Sentilla Corporation
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Sentilla Corporation ("Confidential Information"). Disclosure of this
 * Confidential Information is prohibited, and its use shall be in
 * compliance with terms of the license agreement with Sentilla Corporation.
 */

package com.sentilla.mote.netview;

import com.sentilla.io.ByteBuffer;
import com.sentilla.mote.clientservices.ClientServicesLib;
import com.sentilla.net.ByteReceiver;
import com.sentilla.net.ByteReceiverDriver;
import com.sentilla.system.Leds;
import com.sentilla.system.LedsDriver;
import com.sentilla.system.PropertyDriver;
import com.sentilla.net.Sender;
import com.sentilla.net.SenderDriver;


/**
 * Modified version of ReportHoodRssi that invokes client services heartbeat directly.
 * It thus avoid the memory problem when running clientservices app together with rssi app.
 * This serves as an example of how ClientServices can be componentized.
 * It sends a heartbeat message 5 times less frequently than rssi message.
 *
 *
 * Collects and reports information that helps determine the quality of the wireless
 * sensor network.
 * All messages on port 249 are received. We determine the source mac of each
 * message and save the RSSI in a table.
 * RSSI is summed for a given mac during the reporting interval.
 * A sender MAC is considered to be valid and recognized if it starts at offset 6 of the message
 * and adheres to a MAC pattern currently valid tmotes. That allows all serialized messages
 * that start with mac to be considered for calculations (an example of that is Heartbeat
 * message of ClientServices).
 */
public class RssiCS {



    /*
     * The Report thread receives a beacon broadcast, records its rssi value,
     * then relays that rssi value to its radio neighborhood in a hood message.
    */
    public static class Report {

        /**
         * Blink on certain actions - used for troubleshooting.
         */
        private boolean doBlink = true;

        // used for color indicator of RSSI strength
        /*
        private static final int RSSI_MID_RANGE = 30;
        private static final int RSSI_CLOSE_RANGE = 60;
        */

        /**
         * Port to listen on for traffic. Assuming the long at offset 6 is
         * the mac of the source.
         * Using this approach allows to listen more packets and
         * weed out those whose long starting at offset 6 does not
         * look like a mac.
         */
        private static final short TRAFFIC_PORT = 249;

        /**
         * Size of message that contains a 6 byte header and long mac address
         * minimum valid message: 6 bytes header, 8 bytes src mac
         */
        private static final int MESSAGE_WITH_MAC_SIZE = 14;

        /**
        *
        * Interval used in receive block.
        * A value about 2 s less than report interval introduces
        * convenient variations in receive time between
        * individual motes which improves reporting.
        *
        */
       private static final int BLOCK_INTERVAL = 20000;

        /**
         *
         * Interval between sending RSSI report.
         */
        private static final int REPORT_INTERVAL = 25000;

        /**
         * Time when last report was sent.
         */
        private long lastSent = 0;

        /**
         * Message to be sent to report RSSI.
         */
        private RssiMessage msg;


        public void run() throws InterruptedException {
            Leds leds = LedsDriver.create();
            Sender sender = SenderDriver.create("local");

            // Create an instance of the message to send.
            msg = new RssiMessage();
            msg.id = PropertyDriver.open("mac64").readLong();
            ByteBuffer recvBody = new ByteBuffer(MESSAGE_WITH_MAC_SIZE);
            ByteReceiver trafficReceiver = ByteReceiverDriver.create(TRAFFIC_PORT);
            while (true) {
                recvBody.clear();
                trafficReceiver.setReceive(recvBody).submit();
                int interval = (int) (REPORT_INTERVAL - (System.currentTimeMillis() - lastSent));
                if (interval <= 0) {
                    interval = 10;
                } else if (interval > BLOCK_INTERVAL) {
                    interval = BLOCK_INTERVAL;
                }

                trafficReceiver.block(interval);
                if( trafficReceiver.isDone() ) {
                    msg.packetCount++;
                    short rssi = (short)((trafficReceiver.getRssi() + 64) & 255);
                    long longMac = 0;
                    if (recvBody.available() >= MESSAGE_WITH_MAC_SIZE) {
                        // throw out message header
                        recvBody.readInt();
                        recvBody.readShort();
                        longMac = recvBody.readLong();
                        // unrecognized packet: can't determine src mac address
                        if (isMacAddress(longMac)) {
                            short shortMac = (short) (longMac & 0xffff);
                            // add to statistics
                            short index = msg.ensureEntry(shortMac);
                            short offset = (short) (index * 3);
                            msg.body[offset+1] += rssi;
                            msg.body[offset+2]++;
                        }
                    }
                }
                if (intervalReached()) {
                    resetInterval();
                    if (doBlink) {
                        leds.on(1);
                    }
                    msg.time = System.currentTimeMillis();
                    sender.send(msg);
                    msg.count = 0;
                    if (doBlink) {
                        leds.off(1);
                    }
                    sendHeartbeat();
                }
            }
        }

        /**
         *
         * Send heartbeat on every other invocation.
         * Instantiate ClientServices object
         * if it hasn't been instantiated yet.
         * @throws InterruptedException
         */
        public static void sendHeartbeat() throws InterruptedException {
            if ((heartbeatCounter++ % 2) == 0) {
                if (cs == null) {
                    cs = new ClientServicesLib(true);
                    // don't receive CS commands, just
                    //   send heartbeat
                    cs.receive = false;
                    cs.init();
                }
                cs.loopCount = 1;
                cs.run();
            }
        }

        private static short heartbeatCounter = 0;
        private static ClientServicesLib cs;


        /**
         * Resets the interval.
         *
         */
        private void resetInterval() {
            lastSent = System.currentTimeMillis();
        }

        /**
         * Determines if interval given by current time,
         * lastSent time and the report interval has been reached
         * at this time.
         * @return
         */
        private boolean intervalReached() {
            return (System.currentTimeMillis() - lastSent) >= REPORT_INTERVAL;
        }


    }


    /**
    * Determines if this is a mac address.
     * @param mac long value possibly representing a tmote MAC address
     * @return boolean indicating of the passed value
     *   is a MAC address.
     */
    public static boolean isMacAddress(long mac) {
      /*
       * Sample mac addresses:

mote #2: COM9:  P/HZ: 5195192732659743 0x00127500115EB01F
mote #3: COM6:  P/JA: 519519273 3697467 0x00127500116E85BB
mote #4: COM7:  P/H6: 519519273 2663322 0x00127500115EBE1A

sample mac:
       0x00127500115EBE1A
and mask:
       0xffffffff00000000
result after =& mask:
       0x0012750000000000
       */
      // this is currently the case for all sentilla motes
        return (mac & 0xffffffff00000000L) == 0x0012750000000000L;
    }


    public static void motemain() throws Exception {
        new Report().run();
    }

}

