package com.sentilla.mote.netview;

import java.io.Serializable;

/**
 * Rssi report message that contains statistics about RSSI received from 0 to N motes.
 * The maximum number of entries contains in the body is determined by this formula:
 * floor ((MAX_MESSAGE_SIZE - HEADER_SIZE) / 6),
 * where header size is size of type information plus id,packetCount etc.)
 * <p>
 * Size of this message at MAX_MOTES 8 is 84:
 *n=8
 *array_header=8
 *(6+8+4+8+2)+array_header+n*6
 *
 * </p>
 */
public class RssiMessage implements Serializable {

    /**
     * Maximum number of motes that fits into one report.
     */
    static final short MAX_MOTES = 8;

    /**
     * mac of sender of this message, that is also mac
     * of destination of the message on whose rssi we
     * are reporting.
     */
    public long id;

    /**
     * Total packet count received on port 249 since mote app was installed.
     */
    public int packetCount;

    /**
     * Timestamp of sending report message, in the time
     * system of the sender mote.
     */
    public long time;

    /**
     * Node count in the body.
     */
    public short count;

    /**
     * Body of rssi report, consisting of triples of shorts for each mote's rssi:
     * <p>
     * <pre>
     *  shortMac, rssiSum, count
     * </pre>
     * Where shortMac is the lowest 16 bits of MAC address, rssiSum and count can be used to
     * calculate average of rssi as (rssiSum/count).
     * Size of the body is >= (count * 6).
     * If rssiSum is 0 for an entry, that entry should not be considered as part of valid statistics by
     * the receiving application.
     * </p>
     */
    public short body[] = new short[MAX_MOTES * 3];

    /**
     * Finds the entry with given short Mac
     * @param mac
     * @return 0-based index of entry or -1 if not found
     */
    public short findEntry(short mac) {
        short rc = -1;
        short offset = 0;
        for (short i = 0; i < count; i++) {
            if (body[offset] == mac) {
                rc = i;
                break;
            }
            offset+=3;
        }
        return rc;
    }

    /**
     * Appends entry at end of used portion of body.
     * The caller is responsible for ensuring given
     * entry is not already present, otherwise
     * {@link #findEntry(short)} will ignore
     * entries after the first one.
     * @param shortMac
     */
    public short addEntry(short shortMac) {
        if (count >= MAX_MOTES) {
            throw new IndexOutOfBoundsException("Max mote count exceeded: " + MAX_MOTES);
        }
        int offset = count*3;
        body[offset] = shortMac;
        body[offset+1] = 0;
        body[offset+2] = 0;
        short rc = count;
        count++;
        return rc;
    }

    public short ensureEntry(short mac) {
        short rc;
        rc = findEntry(mac);
        if (rc < 0) {
            rc = addEntry(mac);
        }
        return rc;
    }
}